#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  stdout.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey; monkey.patch_all()
from wishbone.module import OutputModule
from os import getpid
from colorama import init, Fore, Back, Style
import sys
import re
from gevent.fileobject import FileObjectThread


class Format():

    def __init__(self, selection, counter, pid):
        self.selection = selection
        if counter:
            self.counter = self.__returnCounter
        else:
            self.counter = self.__returnNoCounter
        if pid:
            self.pid_value = getpid()
            self.pid = self.__returnPid
        else:
            self.pid = self.__returnNoPid

    def do(self, data):
        return self.pid(self.counter(data))

    def __returnCounter(self, data):
        self.countervalue += 1
        return "%s - %s" % (self.countervalue, data)

    def __returnNoCounter(self, data):
        return data

    def __returnNoPid(self, data):
        return data

    def __returnPid(self, data):
        return "PID-%s: %s" % (self.pid_value, data)


class STDOUT(OutputModule):

    '''
    Prints event data to STDOUT.

    Prints incoming events to STDOUT. When <complete> is True,
    the complete event including headers is printed to STDOUT.

    You can optionally define the colors used.


    Parameters::

        - background_color(str)("RESET")
           |  The background color.
           |  Valid values: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET

        - colorize(bool)(False)
           |  When True all STDOUT output is wrapped in between ANSI color
           |  escape sequences defined by `foreground_color`, `background_color`,
           |  `color_style`.

        - color_style(str)("NORMAL")
           |  The coloring style to use
           |  Valid values: DIM, NORMAL, BRIGHT

        - counter(bool)(False)
           |  Puts an incremental number for each event in front
           |  of each event.

        - foreground_color(str)("WHITE")
           |  The foreground color.
           |  Valid values: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE

        - native_events(bool)(False)
           |  If True, outgoing events are native events.

        - parallel_streams(int)(1)
           |  The number of outgoing parallel data streams.

        - payload(str)(None)
           |  The string to submit.
           |  If defined takes precedence over `selection`.

        - pid(bool)(False)
           |  Includes the pid of the process producing the output.

        - prefix(str)("")*
           |  Puts the prefix in front of each printed event.

        - selection(str)(None)
           |  The event key to submit.
           |  If ``None`` the complete event is selected.


    Queues::

        - inbox
           |  Incoming events.
    '''

    def __init__(self, actor_config,
                 selection=None, payload=None, native_events=False, parallel_streams=1,
                 counter=False, prefix="", pid=False, colorize=False,
                 foreground_color="WHITE", background_color="RESET", color_style="NORMAL"):
        OutputModule.__init__(self, actor_config)

        self.__validateInput(foreground_color, background_color, color_style)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

        self.ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')

    def preHook(self):

        self.format = Format(
            self.kwargs.selection,
            self.kwargs.counter,
            self.kwargs.pid
        )

        if self.kwargs.colorize:
            init(autoreset=True)
            self.getString = self.__stringColor
        else:
            self.getString = self.__stringNoColor

        self.f = FileObjectThread(sys.stdout)

    def consume(self, event):

        data = self.encode(
            self.getDataToSubmit(
                event
            )
        )

        output = self.getString(
            getattr(Fore, event.kwargs.foreground_color),
            getattr(Back, event.kwargs.background_color),
            getattr(Style, event.kwargs.color_style),
            event.kwargs.prefix,
            self.format.do(data)
        )
        self.f.write(output)

    # def postHook(self):

    #     self.f.close()

    def __validateInput(self, f, b, s):

        if f not in ["BLACK", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE"]:
            raise Exception("Foreground value is not correct.")
        if b not in ["BLACK", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE", "RESET"]:
            raise Exception("Background value is not correct.")
        if s not in ["DIM", "NORMAL", "BRIGHT"]:
            raise Exception("Style value is not correct.")

    def __stringColor(self, f, b, s, p, d):
        return "%s%s%s%s%s\n" % (
            f,
            b,
            s,
            p,
            self.format.do(d)
        )

    def __stringNoColor(self, f, b, s, p, d):

        d = self.ansi_escape.sub('', str(d))

        return "%s%s\n" % (
            p,
            self.format.do(d)
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  fresh.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.actor import Actor
from wishbone.module import FlowModule
from gevent import sleep
from wishbone.event import Event


class Fresh(FlowModule):

    '''**Generates a new event unless an event came through in the last x time.**

    This module forwards events without modifying them. If an event has been
    forwarded it resets the timeout counter back to <timeout>.  If the timeout
    counter reaches zero because no messages have passed through, an event
    with <timeout_payload> is generated and submitted to the module's
    <timeout> queue.  When the a timeout_payload has been sent and the event
    stream recovers, a new event with <recovery_payload> is generated and
    submitted to the <timeout> queue.

    Parameters::

        - timeout_payload(int/float/str/obj/list/...)("timeout")
           |  The data a timeout event contains.

        - recovery_payload(int/float/str/obj/list/...)("recovery")
            |  The data a recovery event contains

        - timeout(int)(60)
           |  The max time in seconds allowed to not to receive events.

        - repeat_interval(int)(60)

           |  The interval time to resend the <payload> event in case
           |  <timeout> has expired and

    Queues::

        - inbox
           |  Incoming events.

        - outbox
           |  Outgoing events.

        - timeout
           |  timeout and recovery events.
    '''

    def __init__(self, actor_config, timeout_payload="timeout", recovery_payload="recovery", timeout=60, repeat_interval=60):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("timeout")
        self.registerConsumer(self.consume, "inbox")
        self._counter = self.kwargs.timeout
        self._incoming = False

    def preHook(self):

        self.sendToBackground(self.countDown)

    def consume(self, event):

        self.submit(event, "outbox")
        self._resetTimeout()

    def countDown(self):

        while self.loop():
            if self._counter > 0:
                self._counter -= 1
                sleep(1)
            else:
                self.logging.info("Timeout of %s seconds expired.  Generated timeout event." % (self.kwargs.timeout))
                self._incoming = False
                while self.loop() and not self._incoming:
                    e = Event()
                    e.set(self.kwargs.timeout_payload)
                    self.submit(e, "timeout")
                    self._sleeper(self.kwargs.repeat_interval)
                self.logging.info("Incoming data resumed. Sending recovery event.")
                e = Event()
                e.set(self.kwargs.recovery_payload)
                self.submit(e, "timeout")

    def _resetTimeout(self):

        self._counter = self.kwargs.timeout
        self._incoming = True

    def _sleeper(self, seconds):

        while self.loop() and seconds > 0 and not self._incoming:
            sleep(1)
            seconds -= 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       dictgenerator.py
#
#       Copyright 2016 Jelle Smet development@smetj.net
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#
from sys import version_info
from random import choice, randint
from gevent import sleep
from wishbone import Actor
from wishbone.event import Event
import os


if version_info[0] == 2:
    from io import open


class DictGenerator(Actor):

    '''**Generates random dictionaries.**

    This module allows you to generate an stream of dictionaries.

    Parameters::

        - keys(list)([])
           |  If provided, documents are created using the provided
           |  keys to which random values will be assigned.

        - randomize_keys(bool)(True)
           |  Randomizes the keys.  Otherwise keys are sequential
           |  numbers.

        - num_values(bool)(False)
           |  If true values will be numeric and randomized.

        - num_values_max(int)(1)
           |  The maximum of a value when they are numeric.

        - min_elements(int)(1)
           |  The minimum number of elements per dictionary.

        - max_elements(int)(1)
           |  The maximum number of elements per dictionary.

        - interval(int)(1)
           |  The time in seconds to sleep between each message.

    Queues::

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, keys=[], randomize_keys=True, num_values=False, num_values_min=0, num_values_max=1, min_elements=1, max_elements=1, interval=1):
        Actor.__init__(self, actor_config)

        self.wordlist = self.readWordlist()
        self._total = 0

        self.key_number = -1

        if self.kwargs.randomize_keys:
            self.generateKey = self.pickWord
        else:
            self.generateKey = self.generateKeyNumber

        if self.kwargs.num_values:
            self.generateValue = self.generateValueInteger
        else:
            self.generateValue = self.pickWord

        self.pool.createQueue("outbox")

    def readWordlist(self):
        with open("%s/../data/wordlist.txt" % (os.path.dirname(__file__)), encoding="latin-1") as f:
            return f.readlines()

    def preHook(self):

        if self.kwargs.keys != []:
            self.getDict = self.getDictPredefinedKeys
        else:
            self.getDict = self.getDictGeneratedKeys

        self.sendToBackground(self.generateDicts)

    def generateDicts(self):

        while self.loop():
            d = self.getDict()
            event = Event(d)
            self.submit(event, self.pool.queue.outbox)
            self.key_number = +1
            sleep(self.kwargs.interval)

    def getDictPredefinedKeys(self):

        d = {}
        for key in self.kwargs.keys:
            d[key] = self.pickWord()

        return d

    def getDictGeneratedKeys(self):

        d = {}
        for x in range(0, randint(self.kwargs.min_elements, self.kwargs.max_elements)):
            d[self.generateKey()] = self.generateValue()
        return d

    def pickWord(self):
        '''Returns a word as string from the wordlist.'''

        while self.loop():
            word = choice(self.wordlist).rstrip()
            try:
                return word.encode("ascii", "ignore")
            except:
                pass

    def generateValueInteger(self):
        '''Returns a random number.'''
        return randint(self.kwargs.num_values_min, self.kwargs.num_values_max)

    def generateKeyNumber(self):
        '''Generates a key by incrementing integer.'''
        self.key_number += 1
        return str(self.key_number)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  regex.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.function.template import TemplateFunction
import re


class Regex(TemplateFunction):

    '''
    Regex matching on a string.


    A wishbone template function to do regex matching on strings.  Useful for
    conditional statements.

    Args:
        None
    '''

    def get(self, pattern, string):
        '''
        The function mapped to the template function.

        Args:
            pattern (str): The regex to apply.
            string (str): The data to evaluate.

        Returns:
            Bool: True if regex matches if not False
        '''

        if re.match(pattern, string):
            return True
        else:
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  componentmanager.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import pkg_resources
from prettytable import PrettyTable
from wishbone.error import InvalidComponent, NoSuchComponent
from wishbone.actor import Actor
from wishbone.function.template import TemplateFunction
from wishbone.function.module import ModuleFunction
from wishbone.protocol import Encode
from wishbone.protocol import Decode


class ComponentManager():

    '''
    Loads Wishbone components and information.

    When initiated it indexes all the Wishbone components found in the
    available ``<namespace>.<component_type>.<category>`` combinations.

    A complete component reference would then be something like:
    ``wishbone.module.process.modify``

    - ``wishbone`` is the namespace
    - ``module`` is the component type
    - ``process`` is the component category
    - ``modify`` is the name of the component

    Note:

        The default Wishbone namespaces are ``wishbone``, for the builtin
        modules and ``wishbone_contrib`` or ``wishbone_external`` for
        externally developed components.

        There exist 3 component types:
            - module
            - function
            - protocol

    Args:

        namespace (list): The list of namespaces to search for <categories>
        module_categories (list): The list of module categories to search
        function_categories (list): The list of function categories to search
    '''

    COMPONENT_TYPES = [
        "protocol",
        "module",
        "function",
    ]

    def __init__(self,
                 namespace=["wishbone", "wishbone_contrib", "wishbone_external"],
                 protocol_categories=["encode", "decode"],
                 module_categories=["flow", "input", "output", "process"],
                 function_categories=["template", "module"],
                 ):
        self.namespace = namespace
        self.component_types = ["protocol", "module", "function"]
        self.protocol_categories = protocol_categories
        self.module_categories = module_categories
        self.function_categories = function_categories

    def exists(self, name):

        '''
        Validates whether the component with <name> exists.

        Args:
            name (str): The complete name of the component.

        Returns:
            bool: True if component exists. False otherwise.
        '''

        if self.getComponentByName(name) is None:
            return False
        else:
            return True

    def getComponent(self, namespace, component_type, category, name):
        '''
        Returns the module with name <namespace>.<component_type>.<category>.<name>

        <namespace>.<component_type>.<category>.<name> must be an entrypoint.

        Args:
            namespace (str): The component namespace
            component_type (str): The component type.
            category (str): The component category.
            name (str): The component name.

        Returns:
            class: A ``wishbone.Actor``, ``wishbone.Function`` based class

        Raises:
            NoSuchComponent: The module does not exist.
            InvalidComponent: There was module found but it was not deemed valid.
        '''

        m = None
        for module in pkg_resources.iter_entry_points("%s.%s.%s" % (namespace, component_type, category)):
            if module.name == name:
                m = module.load()
                break

        if m is None:
            raise NoSuchComponent("Component %s.%s.%s.%s cannot be found." % (namespace, component_type, category, name))
        else:
            if issubclass(m, Actor) or issubclass(m, TemplateFunction) or issubclass(m, ModuleFunction) or issubclass(m, Decode) or issubclass(m, Encode):
                return m
            else:
                raise InvalidComponent("'%s.%s.%s.%s' is not a valid wishbone component." % (namespace, component_type, category, name))

    def getComponentByName(self, name):
        '''
        Returns the module with name ``name``.
        ``name`` should be a valid entrypoint.

        Args:
            name (str): The complete module name.

        Returns:
            class: A `wishbone.Actor` or `wishbone.Function` based class

        Raises:
            NoSuchComponent: The module does not exist.
            InvalidComponent: There was module found but it was not deemed valid.
        '''
        self.validateComponentName(name)
        (namespace, component_type, category, name) = name.split('.')
        return self.getComponent(namespace, component_type, category, name)

    def getComponentList(self):
        '''
        Finds and lists all the components found at the defined
        <namespace>.<module_categories>. combinations.

        Yields:
            tuple: A 4 element tuple: (`namespace`, `component_type`, `category`, `name`)
        '''

        for namespace in self.namespace:
            for category in sorted(self.protocol_categories):
                prefix = "%s.protocol.%s" % (namespace, category)
                for item in [m.name for m in pkg_resources.iter_entry_points(group=prefix)]:
                    yield (namespace, "protocol", category, item)
            for category in sorted(self.function_categories):
                prefix = "%s.function.%s" % (namespace, category)
                for item in [m.name for m in pkg_resources.iter_entry_points(group=prefix)]:
                    yield (namespace, "function", category, item)
            for category in sorted(self.module_categories):
                prefix = "%s.module.%s" % (namespace, category)
                for item in [m.name for m in pkg_resources.iter_entry_points(group=prefix)]:
                    yield (namespace, "module", category, item)

    def getComponentDoc(self, namespace, component_type, category, name):
        '''
        Returns the docstring of module `namespace`.`category`.`group`.`name`

        Args:
            namespace (str): The namespace value.
            component_type (str): The component type.
            category (str): The component type category.
            name (str): The component name name.

        Returns:
            str: The docstring of the module.

        Raises:
            InvalidModule: The docstring does not have the correct format.
        '''

        doc = self.getComponent(namespace, component_type, category, name).__doc__

        if doc is None:
            raise InvalidComponent("Component '%s' does not seem to have the expected docstring format." % ())
        else:
            return doc

    def getComponentTitle(self, namespace, component_type, category, name):
        '''
        Returns the title of the module `category`.`group`.`name` docstring.

        Args:
            namespace (str): The namespace value.
            component_type (str): The component type.
            category (str): The component type category.
            name (str): The component name name.

        Returns:
            str: The docstring/module title

        Raises:
            InvalidModule: The docstring does not have the correct format.
        '''

        doc = self.getComponent(namespace, component_type, category, name).__doc__
        if doc is None:
            return "The component doesn't have a docstring."

        title = doc.strip().split('\n')[0].strip('*')
        if title is None:
            raise InvalidComponent("Component '%s' does not seem to have the expected docstring format." % (name))
        else:
            return title

    def getComponentTable(self):
        '''
        Returns an ascii table of all found Wishbone components.

        Returns:
            str: The ascii table containing all modules.
        '''

        table = self.__getComponentTable()
        namespace_header = None
        component_type_header = None
        category_header = None

        for (namespace, component_type, category, name) in self.getComponentList():

            title = self.getComponentTitle(namespace, component_type, category, name)
            version = self.getComponentVersion(namespace, component_type, category, name)

            if namespace_header == namespace:
                namespace = ""
            else:
                namespace_header = namespace
                component_type_header = None
                category_header = None

            if component_type_header == component_type:
                component_type = ""
            else:
                component_type_header = component_type

            if category_header == category:
                category = ""
            else:
                category_header = category
                table.add_row(["", "", "", "", "", ""])

            table.add_row([namespace, component_type, category, name, version, title])

        table.add_row(["", "", "", "", "", ""])
        return table.get_string()

    def getComponentVersion(self, namespace, component_type, category, name):
        '''
        Returns the version of the module.

        Args:
            namespace (str): The namespace value.
            component_type (str): The component type.
            category (str): The component type category.
            name (str): The component name name.

        Returns:
            str: The version of the module.
        '''

        try:
            for module in pkg_resources.iter_entry_points("%s.%s.%s" % (namespace, component_type, category)):
                if module.name == name:
                    return module.dist.version
        except:
            return "?"

    def validateComponentName(self, name):
        '''
        Validates a component reference name for the proper format.

        Args:
            name (str): The name to validate.

        Returns:
            bool: True when valid.  False when invalid.
        '''

        if len(name.split('.')) != 4:
            raise InvalidComponent("Component name '%s' should consist out of 4 parts." % (name))
        if name.split('.')[1] not in self.COMPONENT_TYPES:
            raise InvalidComponent("Component name '%s' has an invalid component type name." % (name))

        return True

    def __getComponentTable(self):
        '''
        Returns a skeleton ascii module table object
        '''

        t = PrettyTable(["Namespace", "Component type", "Category", "Name", "Version", "Description"])
        t.align["Namespace"] = "l"
        t.align["Component type"] = "l"
        t.align["Category"] = "l"
        t.align["Name"] = "l"
        t.align["Version"] = "r"
        t.align["Description"] = "l"
        return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  schema.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


SCHEMA = {
    "type": "object",
    "properties": {
        "protocols": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "protocol": {
                            "type": "string"
                        },
                        "arguments": {
                            "type": "object"
                        }
                    },
                    "required": ["protocol"],
                    "additionalProperties": False,
                },
            },
        },
        "module_functions": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "function": {
                            "type": "string"
                        },
                        "arguments": {
                            "type": "object"
                        }
                    },
                    "required": ["function"],
                    "additionalProperties": False
                }
            }
        },
        "template_functions": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "function": {
                            "type": "string"
                        },
                        "arguments": {
                            "type": "object"
                        }
                    },
                    "required": ["function"],
                    "additionalProperties": False
                }
            }
        },
        "modules": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "module": {
                            "type": "string"
                        },
                        "protocol": {
                            "type": "string"
                        },
                        "description": {
                            "type": "string"
                        },
                        "arguments": {
                            "type": "object"
                        },
                        "functions": {
                            "type": "object"
                        },
                        "event": {
                            "type": "boolean"
                        }
                    },
                    "required": ["module"],
                    "additionalProperties": False
                }
            }
        },
        "routingtable": {
            "type": "array"
        }
    },
    "required": ["modules", "routingtable"],
    "additionalProperties": False
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import os
from gevent import sleep
from wishbone.error import ModuleNotReady
from .structured_data_file import StructuredDataFile


class ModuleConnectionMock(object):

    '''
    A class which can be initially assigned to a variable which is going to
    hold a future connection related object/method. If the actual connection
    to the upstream server has not been finished yet, we have at least a
    predictable error.
    '''

    def __init__(self, message="Not connected"):

        self.message = message

    def __getattr__(self, name):

        raise ModuleNotReady(self.message)


class PIDFile():

    '''
    Handles all PIDfile interactions.
    '''

    def __init__(self, location):
        self.location = location

    def alive(self):
        '''Returns True if at least 1 PID in pidfile is alive.'''

        for pid in self.__readPIDFile():
            if self.__isAlive(pid):
                return True
        return False

    def create(self, pids=[]):
        '''Creates the pidfile containing the provided list of pids.'''

        if self.exists():
            if self.valid():
                raise Exception("Pidfile exists and at least one process running.")
            else:
                self.__deletePIDFile()

        self.__writePIDFile(pids)

    def cleanup(self):
        '''Deleted PID file if possible.'''
        if self.exists():
            self.__deletePIDFile()

    def exists(self):
        '''Returns True if file exists.'''

        return os.path.isfile(self.location)

    def read(self):
        '''Returns the content of the pidfile'''

        return self.__readPIDFile()

    def sendSigint(self, pid):
        '''Sends sigint to PID.'''

        try:
            os.kill(int(pid), 2)
        except:
            pass

        while self.__isAlive(pid):
            sleep(1)

    def valid(self):
        '''Returns True at least one PID within pidfile is still alive.'''

        try:
            for pid in self.__readPIDFile():
                if self.__isAlive(pid):
                    return True
            return False
        except Exception:
            return False

    def __isAlive(self, pid):
        '''Verifies whether pid is still alive.'''

        try:
            os.kill(int(pid), 0)
            return True
        except:
            False

    def __writePIDFile(self, pids=[]):
        '''Writes a list pids to the file.'''

        with open(self.location, 'w') as pidfile:
            for pid in pids:
                pidfile.write("%s\n" % (pid))

    def __readPIDFile(self):
        '''Returns a list of pids values the file contains.'''

        with open(self.location, 'r') as pidfile:
            pids = pidfile.readlines()

        return [int(x) for x in pids]

    def __deletePIDFile(self):
        '''Unconditionally deletes the pidfile.'''

        if os.path.isfile(self.location):
            os.remove(self.location)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  msgpack.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone.protocol import Decode
from msgpack import Unpacker
from msgpack.exceptions import BufferFull
from wishbone.error import ProtocolError


class MSGPack(Decode):

    '''**Decode MSGpack data into a Python data structure.**

    Convert a MSGPack bytestring into a Python data structure using the
    defined charset.

    Parameters:

        - charset(string)("utf-8")
           |  The charset to use to decode the bytestring data.

        - buffer_size(int)(4096)
           |  The max amount of bytes allowed to read for 1 event

    '''

    def __init__(self, charset="utf-8", buffer_size=4096):

        self.charset = charset
        self.buffer_size = buffer_size
        self.unpacker = Unpacker(encoding=self.charset, max_buffer_size=buffer_size)
        self.handle_buffer_size = True

    def handleBytes(self, data):

        try:
            self.unpacker.feed(data)
            for value in self.unpacker:
                if value:
                    yield value
                else:
                    return
                    yield
        except BufferFull:
            self.unpacker = Unpacker(encoding=self.charset, max_buffer_size=self.buffer_size)
            raise ProtocolError("Buffer of %s bytes full." % (self.buffer_size))

    def handleReadLinesMethod(self, data):

        for item in data.readlines():
            for result in self.handler(item):
                yield result

    def flush(self):
        result = next(self.unpacker)
        if result is None:
            return
            yield
        else:
            yield result

        self.unpacker = Unpacker(encoding=self.charset, max_buffer_size=self.buffer_size)



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  error.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


class QueueLocked(Exception):
    pass


class QueueEmpty(Exception):
    pass


class QueueFull(Exception):
    pass


class QueueMissing(Exception):
    pass


class QueueOccupied(Exception):
    pass


class QueueConnected(Exception):
    pass


class SetupError(Exception):
    pass


class ReservedName(Exception):
    pass


class ModuleInitFailure(Exception):
    pass


class FunctionInitFailure(Exception):
    pass


class NoSuchModule(Exception):
    pass


class NoSuchComponent(Exception):
    pass


class MissingKey(Exception):
    pass


class MissingNamespace(Exception):
    pass


class BulkFull(Exception):
    pass


class InvalidData(Exception):
    pass


class InvalidComponent(Exception):
    pass


class InvalidModule(Exception):
    pass


class ModuleNotReady(Exception):
    pass


class InvalidConfig(Exception):
    pass


class InvalidEventFormat(Exception):
    pass


class BufferExceeded(Exception):
    pass


class ProtocolError(Exception):
    pass


class ProtocolInitFailure(Exception):
    pass


class TTLExpired(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  setup.py
#
#  Copyright 2017 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

PROJECT = 'wishbone'
VERSION = '3.1.1'

install_requires = [
    'arrow',
    'colorama',
    'cronex',
    'gevent',
    'gipc',
    'inotify>=0.2.9',
    'jsonschema',
    'prettytable',
    'python-daemon-3K',
    'requests',
    'pyyaml',
    'jsonschema',
    'msgpack-python',
    'easydict',
    'jinja2',
    'pygments',
    'scalpl'
]

dependency_links = [
]

try:
    with open('README.rst', 'rt') as f:
        long_description = f.read()
except IOError:
    long_description = ''


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            "-v",
            "tests/"
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


setup(
    name=PROJECT,
    version=VERSION,

    description='Build composable event pipeline servers with minimal effort.',
    long_description=long_description,

    author='Jelle Smet',
    author_email='development@smetj.net',

    url='https://github.com/smetj/wishbone',
    download_url='https://github.com/smetj/wishbone/tarball/master',
    dependency_links=dependency_links,
    classifiers=['Development Status :: 5 - Production/Stable',
                 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: Implementation :: PyPy',
                 'Intended Audience :: Developers',
                 'Intended Audience :: System Administrators',
                 ],
    extras_require={
        'testing': ['pytest'],
    },
    platforms=['Linux'],
    test_suite='tests.test_wishbone',
    cmdclass={'test': PyTest},
    scripts=[],
    provides=[],
    install_requires=install_requires,
    namespace_packages=[],
    packages=find_packages(),
    package_data={'': ['data/wordlist.txt', 'data/LICENCE', 'data/banner.tmpl']},
    zip_safe=False,
    entry_points={
        'console_scripts': ['wishbone = wishbone.bootstrap:main'],
        'wishbone.protocol.decode': [
            'dummy = wishbone.protocol.decode.dummy:Dummy',
            'plain = wishbone.protocol.decode.plain:Plain',
            'json = wishbone.protocol.decode.json:JSON',
            'msgpack = wishbone.protocol.decode.msgpack:MSGPack',
        ],
        'wishbone.protocol.encode': [
            'binary = wishbone.protocol.encode.binary:Binary',
            'dummy = wishbone.protocol.encode.dummy:Dummy',
            'json = wishbone.protocol.encode.json:JSON',
            'msgpack = wishbone.protocol.encode.msgpack:MSGPack',
        ],
        'wishbone.module.flow': [
            'acknowledge = wishbone.module.acknowledge:Acknowledge',
            'count = wishbone.module.count:Count',
            'fanout = wishbone.module.fanout:Fanout',
            'funnel = wishbone.module.funnel:Funnel',
            'fresh = wishbone.module.fresh:Fresh',
            'queueselect = wishbone.module.queueselect:QueueSelect',
            'roundrobin = wishbone.module.roundrobin:RoundRobin',
            'switch = wishbone.module.switch:Switch'

        ],
        'wishbone.module.process': [
            'modify = wishbone.module.modify:Modify',
            'pack = wishbone.module.pack:Pack',
            'template = wishbone.module.template:Template',
            'unpack = wishbone.module.unpack:Unpack',
        ],
        'wishbone.module.input': [
            'cron =  wishbone.module.cron:Cron',
            'inotify = wishbone.module.wb_inotify:WBInotify',
            'generator = wishbone.module.generator:Generator',
        ],
        'wishbone.module.output': [
            'null = wishbone.module.null:Null',
            'stdout = wishbone.module.stdout:STDOUT',
            'syslog = wishbone.module.wbsyslog:Syslog',
            'throughput = wishbone.module.throughput:Throughput'
        ],
        'wishbone.function.module': [
            'set = wishbone.function.module.set:Set',
            'append = wishbone.function.module.append:Append',
            'uppercase = wishbone.function.module.uppercase:Uppercase',
            'lowercase = wishbone.function.module.lowercase:Lowercase',
        ],
        'wishbone.function.template': [
            'choice = wishbone.function.template.choice:Choice',
            'cycle = wishbone.function.template.cycle:Cycle',
            'environment = wishbone.function.template.environment:Environment',
            'epoch = wishbone.function.template.epoch:Epoch',
            'pid = wishbone.function.template.pid:PID',
            'random_bool = wishbone.function.template.random_bool:RandomBool',
            'random_integer = wishbone.function.template.random_integer:RandomInteger',
            'random_uuid = wishbone.function.template.random_uuid:RandomUUID',
            'random_word = wishbone.function.template.random_word:RandomWord',
            'regex = wishbone.function.template.regex:Regex',
            'strftime = wishbone.function.template.strftime:STRFTime',
            'version = wishbone.function.template.version:Version'
        ]
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  switch.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.actor import Actor
from wishbone.module import FlowModule
from wishbone.error import ModuleInitFailure, ReservedName


class Switch(FlowModule):

    '''**Switch outgoing queues while forwarding events.**

    Forwards events to the desired outgoing queue based on the value of
    <outgoing>.

    The value of <outgoing> can be dynamically set in 2 ways:

        - Using a template function.

        - By sending an event to the <switch> queue with the value of
          <outgoing> stored under *data*.


    Parameters::

        - outgoing(str)("outbox")*
            |  The name of the queue to submit incoming events to.

    Queues::

        - inbox
           |  incoming events

        - switch
           |  incoming events to alter outgoing queue.

        - outbox*
           |  outgoing events

        - <connected_queue_1>
           |  outgoing events

        - <connected_queue_n>
           |  outgoing events
    '''

    def __init__(self, actor_config, outgoing="outbox"):

        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.pool.createQueue("switch")
        self.pool.createQueue("outbox")
        self.pool.queue.switch.disableFallThrough()
        self.registerConsumer(self.consume, "inbox")
        self.registerConsumer(self.switch, "switch")

        self.forbidden = ["inbox", "switch"]

    def preHook(self):

        if self.kwargs.outgoing in self.forbidden:
            raise ModuleInitFailure("Module parameter <outgoing> cannot have value '%s'." % (self.kwargs.outgoing))

        self.destination = self.kwargs.outgoing
        self._destination = self.kwargs.outgoing

    def consume(self, event):

        if self.kwargs.outgoing != self._destination:
            self._destination = self.kwargs.outgoing
            self.destination = self.kwargs.outgoing

        if self.destination in self.forbidden:
            raise ReservedName("Cannot forward incoming events to queue '%s'." % (self.destination))
        else:
            self.submit(event, self.destination)

    def switch(self, event):

        prefix = "<switch> queue received event"
        try:
            name = event.get("data")
            if self.pool.hasQueue(name):
                self.destination = name
                self.logging.info("%s. Outgoing messages forwarded to queue '%s'." % (prefix, name))
            else:
                self.logging.error("%s but module has no queue named '%s'." % (prefix, name))
        except KeyError:
            self.logging.error("%s but has no value key data." % (prefix))



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  queueselect.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import ProcessModule
from wishbone.utils.structured_data_file import StructuredDataFile
from wishbone.error import InvalidData


class QueueSelect(ProcessModule):

    '''**Submits message to the queue defined by a rendered template.**

    Renders a list of templates against the event.  Events are submitted to
    the queue a rendered template returns.

    Typically this module is used to route messages using Jinja2 conditionals
    and logic.

    A rule rule looks like this:

    ::

        { "name": "name of the rule",
          "queue": "{{ 'queue_1,queue_one' if data.one == 1 else 'queue_2,queue_two' }}"
          "payload":
            "queue_1": {
                "detail_1": "some value",
                "detail_2": "some other value",
            },
            "queue_2": {
                "detail_1": "some value",
                "detail_2": "some other value",
            }
          }
        }


    The <file> queue expects events containing the absolute path of a YAML
    file to read (or delete). Typically this queue receives events from
    wishbone.module.input.inotify.

    Events of type "IN_CREATE", "IN_CLOSE_WRITE", "IN_DELETE", "WISHBONE_INIT"
    are processed all others are ignored.


    Parameters::

        - templates(list)([])*
           |  A list consisting out of template dicts as explained above.

        - log_matching(bool)(False)
           |  Whether to produce debug log messages for matches.
           | Can be verbose hence it's configurable.


    Queues::

        - inbox
           |  Incoming events

        - outbox
           |  Outgoing events

        - nomatch
           |  Events not matching at least 1 rule.

        - file
           |  Read rules from YAML file or delete them.

    '''

    INOTIFY_TYPES = [
        "IN_CREATE",
        "IN_CLOSE_WRITE",
        "IN_DELETE",
        "WISHBONE_INIT"
    ]

    def __init__(self, actor_config, templates=[], log_matching=False):
        ProcessModule.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("file")
        self.pool.createQueue("nomatch")

        self.registerConsumer(self.consume, "inbox")
        self.registerConsumer(self.handleFileTemplate, "file")

        self.template_loader = StructuredDataFile(
            expect_json=False,
            expect_yaml=True
        )

    def consume(self, event):

        for template in event.kwargs.templates:
            self.handleQueueSelect(
                template_name=template.name,
                queue_list=template.queue,
                payload=template.get("payload", {}),
                event=event)

        for file_name, file_content in self.template_loader.dump().items():
            try:
                queue_name = event.render(file_content["queue"])
            except InvalidData as err:
                self.logging.error("Failed to render template '%s'. Reason: %s" % (err))
            else:
                self.handleQueueSelect(
                    template_name=file_name,
                    queue_list=queue_name,
                    payload=file_content["payload"].get(queue_name, {}),
                    event=event)

    def handleQueueSelect(self, template_name, queue_list, payload, event):
        '''Handles submitting <event> into queue <queue_name>.'''

        for queue_name in [queue.strip() for queue in queue_list.split(',')]:

            if self.pool.hasQueue(queue_name):

                if self.kwargs.log_matching:
                    self.logging.debug("Template '{template_name}' selected queue '{queue_name}' to route event '{event_id}' to.".format(
                        template_name=template_name,
                        queue_name=queue_name,
                        event_id=event.get('uuid')
                    ))

                # Construct and set the payload
                queue_payload = {
                    "original_event_id": event.get('uuid'),
                    "queue": queue_name,
                    "payload": payload
                }
                e = event.clone()
                e.set(queue_payload, "tmp.%s" % (self.name))

                # Submit a clone of the event to the required queue
                self.submit(e, queue_name)

            else:
                if self.kwargs.log_matching:
                    self.logging.debug("Template '{template_name}' selected non-existing queue '{queue_name}' to route event '{event_id}' to.".format(
                        template_name=template_name,
                        queue_name=queue_name,
                        event_id=event.get('uuid')
                    ))
                self.submit(event, "nomatch")

    def handleFileTemplate(self, event):
        '''Loads or deletes the template file defined in data.path.'''

        inotify_type = event.get("data.inotify_type")
        path = event.get("data.path")

        if inotify_type in self.INOTIFY_TYPES:
            if inotify_type == "IN_DELETE":
                self.template_loader.delete(path)
                self.logging.debug("Removed template file '{path}' from cache.".format(path=path))
            else:
                self.template_loader.load(path)
                self.logging.debug("Loaded template file '{path}'".format(path=path))
        else:
            self.logging.warning("No support for inotify type '{inotify_type}'. Dropped.".format(
                inotify_type=inotify_type
            ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  graphcontent.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import json

GRAPHCONTENT = """
<!doctype html>
<html>
<head>
    <title>Network | Custom Groups</title>

  <style type="text/css">
    # mynetwork {
      width: 100px;
      height: 100px;
      border: 0px solid lightgray;
    }
  </style>

    <script type="text/javascript" src="http://visjs.org/dist/vis.js"></script>
    <link href="http://visjs.org/dist/vis.css" rel="stylesheet" type="text/css"/>
    <link rel="stylesheet" href="http://maxcdn.bootstrapcdn.com/font-awesome/4.3.0/css/font-awesome.min.css">
</head>

<body>

<div id="mynetwork"></div>

<script type="text/javascript">
  // create an array with nodes
  var nodes = %s;

  // create an array with edges
  var edges = %s;

  // create a network
  var container = document.getElementById('mynetwork');
  var data = {
    nodes: nodes,
    edges: edges
  };

//    var options = {
//        "height":  window.innerHeight,
//        "width":  window.innerWidth,
//        "edges": {
//            "smooth": {
//              "forceDirection": "none"
//            }
//        }
//    };
var options = {
  "height":  window.innerHeight,
  "width":  window.innerWidth,
}
  var network = new vis.Network(container, data, options);
</script>

</body>
</html>

"""


class VisJSData():

    def __init__(self, ignore_logging=False, ingore_metrics=False):

        self.id = 0
        self.nodes = {}
        self.edges = []

    def addModule(self, instance_name, module_name, description):

        if ".input." in module_name:
            icon = {
                "face": "FontAwesome",
                "code": '\uf090',
                "size": 75,
                "color": "red"
            }

        elif ".output." in module_name:
            icon = {
                "face": "FontAwesome",
                "code": '\uf08b',
                "size": 75,
                "color": "green"
            }

        elif ".flow." in module_name:
            icon = {
                "face": "FontAwesome",
                "code": '\uf074',
                "size": 75,
                "color": "black"
            }
        else:
            icon = {
                "face": "FontAwesome",
                "code": '\uf085',
                "size": 75,
                "color": "black"
            }

        if instance_name not in list(self.nodes.keys()):
            self.nodes[instance_name] = {"id": self.id,
                                         "label": instance_name,
                                         "shape": "icon",
                                         "icon": icon,
                                         "size": 20,
                                         "title": "Module: %s</br>Description: %s" % (module_name, description),
                                         "font.size": 20
                                         }
            self.id += 1

    def addQueue(self, module, name):

        if "%s.%s" % (module, name) not in list(self.nodes.keys()):
            self.nodes["%s.%s" % (module, name)] = {"id": self.id,
                                                    "label": name,
                                                    "shape": "icon",
                                                    "icon": {
                                                        "face": "FontAwesome",
                                                        "code": '\uf0c9',
                                                        "size": 20,
                                                        "color": "orange"
                                                    },
                                                    "size": 5,
                                                    "title": "Queue: %s.%s" % (module, name),
                                                    "font.size": 20
                                                    }
            self.edges.append({"from": self.nodes[module]["id"],
                               "to": self.nodes["%s.%s" % (module, name)]["id"],
                               "dashes": True})
            self.id += 1

    def addEdge(self, f, t):

        self.edges.append({"from": self.nodes[f]["id"],
                           "to": self.nodes[t]["id"],
                           "arrows": "to"
                           }
                          )

    def dumpString(self):

        nodes = json.dumps(list(self.nodes.values()))
        edges = json.dumps(self.edges)
        return (nodes, edges)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  count.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import FlowModule
from gevent.pool import Pool
from gevent import sleep


class Count(FlowModule):

    '''**Pass or drop events based on the number of times an event value occurs.**

    Events pass through or get dropped after a certain key/value has appeared
    for a number of times within an optional time window.

    When the time window expires, the occurance counter for that field is
    reset.

    Conditions have following format:

    Example 1::

        {
        "data": {
            "value": "abc",
            "occurrence": 10,
            "window": 60
            "action": "pass"
            }
        }

    This means if an event of which the <data> field has value "abc" occurs 10
    times within a time window of 60 seconds since the first occurance
    happened, the 10th event will pass through. The first 9 events will get
    dropped.


    Example 2::

        {
        "tmp.address": {
            "value": "127.0.0.1",
            "occurence": 10,
            "window": 60,
            "action": "drop"
            }
        }

    This means that events with field <tmp.address> and value <127.0.0.1> can
    pass through 10 times after which the events get dropped within a time
    window of 60 seconds


    A window of `0 seconds` disables the time window expiration for a key.

    Events which do not have the requested key can pass through.

    Parameters::

        - conditions(dict)({})
           |  The conditions which should be met.

        - expire(int)(0)
           |  The time window in which all occurances should happen.
           |  A value of 0 disables the expiration.


    Queues::

        - inbox
           |  Incoming events

        - outbox
           |  Outgoing events

        - dropped
           |  Events not meeting the counter requirements.

    '''

    def __init__(self, actor_config, conditions={}):
        FlowModule.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("dropped")
        self.registerConsumer(self.consume, "inbox")

        self.__counter = {}
        self.__count_down_pool = Pool(len(self.kwargs.conditions))

    def consume(self, event):

        for key, condition in event.kwargs.conditions.items():
            if event.has(key):
                if event.get(key) == condition["value"]:
                    if key in self.__counter:
                        self.__counter[key] += 1
                    else:
                        self.__counter[key] = 1
                        if condition["window"] > 0:
                            self.__count_down_pool.spawn(self.__countDown, condition["window"], key)

                    if self.__counter[key] >= condition["occurrence"]:
                        if condition["action"] == "pass":
                            self.submit(event, "outbox")
                        if condition["action"] == "drop":
                            self.submit(event, "dropped")
                            self.logging.debug("Event with id '%s' and key '%s' dropped" % (event.get('uuid'), key))
                    else:
                        if condition["action"] == "pass":
                            self.submit(event, "dropped")
                            self.logging.debug("Event with id '%s' and key '%s' dropped" % (event.get('uuid'), key))
                        if condition["action"] == "drop":
                            self.submit(event, "outbox")
                    break
        else:
            self.submit(event, "outbox")
            self.logging.debug("Event with id '%s' has not a single key defined in the conditions therefor it is passed to outbox." % (event.get('uuid'), key))

    def __countDown(self, seconds, key):

        sleep(seconds)
        del(self.__counter[key])
        self.logging.debug("Time window of '%s' expired for key '%s'." % (seconds, key))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  cycle.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.function.template import TemplateFunction
from itertools import cycle as cycle_array


class Cycle(TemplateFunction):
    '''
    Cycles through the provided array returning the next element.

    A Wishbone template function which returns the next value of an array.

    This function rotates through the elements in the provided array always
    returning the next element.  The order is fixed and when the end is
    reached the first element is returned again.

    Args:

        values(list): A list of elements to cycle through.
    '''

    def __init__(self, values):

        self.c = cycle_array(values)

    def get(self):
        '''
        The function mapped to the template function.

        Args:
            None

        Returns:
            obj: An element of the provided array.
        '''

        return next(self.c)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  unpack.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import ProcessModule
from wishbone.actor import Actor
from wishbone.event import Event
from wishbone.error import InvalidData


class Unpack(ProcessModule):

    '''**Unpacks bulk events into single events.**

    Creates single events from all the events stored in a bulk event.


    Parameters::

        None

    Queues::

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messages

        - dropped
           |  Dropped messages
    '''

    def __init__(self, actor_config):

        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("dropped")
        self.registerConsumer(self.consume, "inbox")

    def consume(self, event):
        if event.isBulk():
            for e in event.dump()["data"]:
                try:
                    self.submit(Event().slurp(e), "outbox")
                except InvalidData:
                    self.logging.debug("Bulk event with id '%s' contained an invalid event. Invalid event skipped." % (event.get('uuid')))
            self.logging.debug("Expanded Bulk event into %s events." % (len(event.data)))
        else:
            self.logging.debug("Event with id '%s' is not a bulk event. Dropped.")
            self.submit(event, "dropped")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  tippingbucket.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Bulk
from wishbone.error import BulkFull
from gevent import sleep


class Bucket(object):

    def __init__(self, key, size, age, logging, queue, looplock):

        self.key = key
        self.size = size
        self.age = age
        self.logging = logging
        self.queue = queue
        self.loop = looplock

        self.bucket = None
        self.createEmptyBucket()
        self.logging.info("Created new bucket with aggregation key '%s'." % (self.key))

    def createEmptyBucket(self):
        self.bucket = Bulk(self.size)
        self.resetTimer()

    def flushBucketTimer(self):

        '''
        Flushes the buffer when <bucket_age> has expired.
        '''

        while self.loop():
            sleep(1)
            self._timer -= 1
            if self._timer == 0:
                if self.bucket.size() > 0:
                    self.logging.debug("Bucket age expired after %s s." % (self.age))
                    self.flush()
                else:
                    self.resetTimer()

    def flush(self):
        '''
        Flushes the buffer.
        '''
        self.logging.debug("Flushed bucket '%s' of size '%s'" % (self.key, self.bucket.size()))
        self.queue.put(self.bucket)
        self.createEmptyBucket()

    def resetTimer(self):
        '''
        Resets the buffer expiry countdown to its configured value.
        '''

        self._timer = self.age


class TippingBucket(Actor):

    '''**Aggregates multiple events into bulk.**

    Aggregates multiple incoming events into bulk usually prior to submitting
    to an output module.

    Flushing the buffer can be done in various ways:

      - The age of the bucket exceeds <bucket_age>.
      - The size of the bucket reaches <bucket_size>.
      - Any event arrives in queue <flush>.


    Parameters::

        - bucket_size(int)(100)
           |  The maximum amount of events per bucket.

        - bucket_age(int)(10)
           |  The maximum age in seconds before a bucket is closed and
           |  forwarded.  This actually corresponds the time since the first
           |  event was added to the bucket.

        - aggregation_key(str)("default")
           |  Groups events with key <aggregation_key> into the same buckets.

    Queues::

        - inbox
           |  Incoming events

        - outbox
           |  Outgoing bulk events

        - flush
           |  Flushes the buffer on incoming events despite the bulk being
           |  full (bucket_size) or expired (bucket_age).

    '''

    def __init__(self, actor_config, bucket_size=100, bucket_age=10, aggregation_key="default"):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("flush")
        self.registerConsumer(self.consume, "inbox")
        self.registerConsumer(self.flushIncomingMessage, "flush")

        self.buckets = {}

    def consume(self, event):

        try:
            self.getBucket(self.kwargs.aggregation_key).bucket.append(event)
        except BulkFull:
            self.logging.debug("Bucket full after %s events." % (self.kwargs.bucket_size))
            self.getBucket(self.kwargs.aggregation_key).flush()
            self.getBucket(self.kwargs.aggregation_key).bucket.append(event)

    def getBucket(self, key):

        if key in self.buckets:
            return self.buckets[key]
        else:
            self.buckets[key] = Bucket(
                self.kwargs.aggregation_key,
                self.kwargs.bucket_size,
                self.kwargs.bucket_age,
                self.logging,
                self.pool.queue.outbox,
                self.loop)
            self.sendToBackground(self.buckets[key].flushBucketTimer)
            return self.buckets[key]

    def flushIncomingMessage(self, event):
        '''
        Called on each incoming messages of <flush> queue.
        Flushes the buffer.
        '''

        self.logging.debug("Recieved message in <flush> queue.  Flushing all bulk buffers.")
        for index, bucket in self.buckets:
            bucket.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  wbsyslog.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import OutputModule
import syslog
import sys
import os


class Syslog(OutputModule):

    '''**Submits event data to syslog.**

    Logevents have following format:

    (6, 1367682301.430527, 'Router', 'Received SIGINT. Shutting down.')

    The first value corresponds to the syslog severity level.

    Parameters::

        - ident(str)(<script_name>)*
           |  The syslog id string.
           |  If not provided the script name is used.
           |  (Can be a dynamic value)

        - level(int)(5)*
           |  The loglevel.
           |  (Can be a dynamic value)

        - native_events(bool)(False)
           |  If True, outgoing events are native events.

        - payload(str)(None)
           |  The string to submit.
           |  If defined takes precedence over `selection`.

        - parallel_streams(int)(1)
           |  The number of outgoing parallel data streams.

        - selection(str)("data")
           |  The event key to submit.

    Queues::

        - inbox
           |  incoming events
    '''

    def __init__(self, actor_config,
                 selection="data", payload=None, native_events=False, parallel_streams=1,
                 level=5, ident=os.path.basename(sys.argv[0])):
        OutputModule.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        syslog.openlog("%s[%s]" % (self.kwargs.ident, os.getpid()))

    def consume(self, event):

        data = self.encode(
            self.getDataToSubmit(
                event
            )
        )

        syslog.syslog(
            event.kwargs.level,
            data
        )

    def postHook(self):
        syslog.closelog()
